#!/usr/bin/env python
# coding: utf-8

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from os import getenv, makedirs
from os.path import exists as path_exists, isfile, join as path_join

from py2neo.internal.compat import urlretrieve
from py2neo.internal.versioning import Version


editions = [
    "community",
    "enterprise",
]

versions = [

    # 3.0 series
    "3.0.0", "3.0.1", "3.0.2", "3.0.3", "3.0.4", "3.0.5", "3.0.6", "3.0.7", "3.0.8", "3.0.9",
    "3.0.10", "3.0.11", "3.0.12",

    # 3.1 series
    "3.1.0", "3.1.1", "3.1.2", "3.1.3", "3.1.4", "3.1.5", "3.1.6", "3.1.7", "3.1.8", "3.1.9",

    # 3.2 series
    "3.2.0", "3.2.1", "3.2.2", "3.2.3", "3.2.5", "3.2.6", "3.2.7", "3.2.8", "3.2.9", "3.2.10",
    "3.2.11", "3.2.12",

    # 3.3 series
    "3.3.0", "3.3.1", "3.3.2", "3.3.3", "3.3.4", "3.3.5", "3.3.6",

    # 3.4 series
    "3.4.0", "3.4.1",

]

# TODO: make this a bit easier to read
version_tuples = [Version(v) for v in versions]
minor_version_tuples = sorted({(str(v.major), str(v.minor)) for v in version_tuples})
minor_versions = [".".join(map(str, v)) for v in minor_version_tuples]
latest_version_tuples = {w: sorted(v for v in version_tuples if v[:2] == w)[-1] for w in minor_version_tuples}
latest_versions = {".".join(map(str, k)): str(v) for k, v in latest_version_tuples.items()}
version_aliases = dict(latest_versions, **{k + "-LATEST": v for k, v in latest_versions.items()})
version_aliases["LATEST"] = versions[-1]

dist = "http://{}".format(getenv("NEO4J_DIST_HOST") or "dist.neo4j.org")
dist_overrides = {
    # "3.0.0-NIGHTLY": "http://alpha.neohq.net/dist",
}

archive_format = getenv("ARCHIVE_FORMAT") or "gz"


class Distribution(object):
    """ Represents a Neo4j archive.
    """

    def __init__(self, edition=None, version=None):
        edition = edition.lower() if edition else "community"
        if edition in editions:
            self.edition = edition
        else:
            raise ValueError("Unknown edition %r" % edition)
        version = version.upper() if version else "LATEST"
        self.snapshot = "SNAPSHOT" in version
        if version in version_aliases:
            version = version_aliases[version]
        if version in versions:
            self.version = version
        else:
            raise ValueError("Unknown version %r" % version)

    @property
    def key(self):
        """ The unique key that identifies the archive, e.g.
        ``community-2.3.2``.
        """
        return "%s-%s" % (self.edition, self.version)

    @property
    def name(self):
        """ The full name of the archive file, e.g.
        ``neo4j-community-2.3.2-unix.tar.gz``.
        """
        return "neo4j-{}-unix.tar.{}".format(self.key, archive_format)

    @property
    def uri(self):
        """ The URI from which this archive may be downloaded, e.g.
        ``http://dist.neo4j.org/neo4j-community-2.3.2-unix.tar.gz``.
        """
        if self.version in dist_overrides:
            return "%s/%s" % (dist_overrides[self.version], self.name)
        else:
            return "%s/%s" % (dist, self.name)

    def download(self, path=".", overwrite=False):
        """ Download a Neo4j distribution to the specified path.

        :param path:
        :param overwrite:
        :return: the name of the downloaded file
        """
        file_name = path_join(path, self.name)
        if overwrite:
            if path_exists(file_name) and not isfile(file_name):
                raise IOError("Cannot overwrite directory %r" % file_name)
        elif not self.snapshot and path_exists(file_name):
            return file_name
        try:
            makedirs(path)
        except OSError:
            pass
        print("Downloading <%s>" % self.uri)
        urlretrieve(self.uri, file_name)
        return file_name



#!/usr/bin/env python
# coding: utf-8

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import click

from py2neo.internal.addressing import NEO4J_URI, NEO4J_AUTH


@click.group(help="""\
Multipurpose Neo4j toolkit.
""")
def py2neo():
    pass


@py2neo.command("version", help="""\
Display the current library version
""")
def py2neo_version():
    from py2neo.meta import __version__
    click.echo(__version__)


@py2neo.command("console", help="""\
Interactive Cypher console
""")
@click.option("-u", "--uri", default=NEO4J_URI, help="Set the connection URI.")
@click.option("-a", "--auth", default=NEO4J_AUTH, help="Set the user and password.")
@click.option("-s", "--secure", is_flag=True, default=False, help="Use encrypted communication (TLS).")
@click.option("-v", "--verbose", is_flag=True, default=False, help="Show low level communication detail.")
def py2neo_console(uri, auth=None, secure=None, verbose=None):
    from py2neo.console import Console, ConsoleError
    try:
        con = Console(uri, auth=auth, secure=secure, verbose=verbose)
    except ConsoleError as error:
        click.echo(error)
        raise SystemExit(1)
    else:
        raise SystemExit(con.loop())


@py2neo.group("auth", help="""\
Auth file management
""")
def py2neo_auth():
    pass


@py2neo_auth.command("list", help="""\
List users
""")
@click.argument("auth_file")
def py2neo_auth_list(auth_file):
    from py2neo.admin.install import AuthFile
    for user in AuthFile(auth_file):
        click.echo(user.name)


@py2neo_auth.command("remove", help="""\
Remove a user

Example:

    py2neo auth remove data/dbms/auth alice
    
""")
@click.argument("auth_file")
@click.argument("user_name")
def py2neo_auth_remove(auth_file, user_name):
    from py2neo.admin.install import AuthFile
    AuthFile(auth_file).remove(user_name)


@py2neo_auth.command("update", help="""\
Create a user or set a password

For general interactive use, password and confirmation prompts will be presented.
For batch mode, use the --password option.

Example:

    py2neo auth update data/dbms/auth alice

If AUTH_FILE contains only a dash `-` then the auth file entry will be written to stdout instead.
""")
@click.argument("auth_file")
@click.argument("user_name")
@click.password_option()
def py2neo_auth_update(auth_file, user_name, password):
    from py2neo.admin.install import AuthFile
    AuthFile(auth_file).update(user_name, password)


@py2neo.command("get", help="""\
Download a Neo4j distribution
""")
def py2neo_get():
    from py2neo.admin.dist import Distribution
    Distribution().download()


@py2neo.command("run", help="""\
Run a Cypher query
""")
@click.option("-u", "--uri", default=NEO4J_URI, help="Set the connection URI.")
@click.option("-a", "--auth", default=NEO4J_AUTH, help="Set the user and password.")
@click.option("-s", "--secure", is_flag=True, default=False, help="Use encrypted communication (TLS).")
@click.option("-v", "--verbose", is_flag=True, default=False, help="Show low level communication detail.")
@click.argument("cypher", nargs=-1)
def py2neo_run(cypher, uri, auth=None, secure=None, verbose=None):
    from py2neo.console import Console, ConsoleError
    try:
        con = Console(uri, auth=auth, secure=secure, verbose=verbose)
    except ConsoleError as error:
        click.echo(error)
        raise SystemExit(1)
    else:
        raise SystemExit(con.run_all(cypher))


def main():
    try:
        py2neo(obj={})
    except Exception as error:
        click.secho(error.args[0], err=True)
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.internal.collections import round_robin


def hydrate_node(graph, identity, inst=None, **rest):
    if inst is None:

        def inst_constructor():
            from py2neo.data import Node
            new_inst = Node()
            new_inst.graph = graph
            new_inst.identity = identity
            new_inst._stale.update({"labels", "properties"})
            return new_inst

        inst = graph.node_cache.update(identity, inst_constructor)
    else:
        inst.graph = graph
        inst.identity = identity
        graph.node_cache.update(identity, inst)

    if "data" in rest:
        inst._stale.discard("properties")
        inst.clear()
        inst.update(rest["data"])
    if "metadata" in rest:
        inst._stale.discard("labels")
        metadata = rest["metadata"]
        inst._remote_labels = frozenset(metadata["labels"])
        inst.clear_labels()
        inst.update_labels(metadata["labels"])
    return inst


def hydrate_relationship(graph, identity, inst=None, **rest):
    start = rest["start"]
    end = rest["end"]

    if inst is None:

        def inst_constructor():
            from py2neo.data import Relationship
            new_inst = Relationship(hydrate_node(graph, start), rest.get("type"),
                                    hydrate_node(graph, end), **rest.get("data", {}))
            new_inst.graph = graph
            new_inst.identity = identity
            return new_inst

        inst = graph.relationship_cache.update(identity, inst_constructor)
    else:
        inst.graph = graph
        inst.identity = identity
        hydrate_node(graph, start, inst=inst.start_node)
        hydrate_node(graph, end, inst=inst.end_node)
        inst._type = rest.get("type")
        if "data" in rest:
            inst.clear()
            inst.update(rest["data"])
        else:
            inst._stale.add("properties")
        graph.relationship_cache.update(identity, inst)
    return inst


def hydrate_path(graph, data):
    from py2neo.data import Path
    node_ids = data["nodes"]
    relationship_ids = data["relationships"]
    offsets = [(0, 1) if direction == "->" else (1, 0) for direction in data["directions"]]
    nodes = [hydrate_node(graph, identity) for identity in node_ids]
    relationships = [hydrate_relationship(graph, identity,
                                          start=node_ids[i + offsets[i][0]],
                                          end=node_ids[i + offsets[i][1]])
                     for i, identity in enumerate(relationship_ids)]
    inst = Path(*round_robin(nodes, relationships))
    inst.__metadata = data
    return inst



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from collections import namedtuple

from neo4j.packstream.structure import Structure
from neo4j.v1.types import PackStreamHydrator as _PackStreamHydrator

from py2neo.internal.hydration import hydrate_node, hydrate_relationship


_unbound_relationship = namedtuple("UnboundRelationship", ["id", "type", "properties"])


class PackStreamHydrator(_PackStreamHydrator):

    def __init__(self, graph, keys, entities=None):
        super(PackStreamHydrator, self).__init__(2)  # maximum known protocol version
        self.graph = graph
        self.keys = keys
        self.entities = entities or {}

    def hydrate(self, values):
        """ Hydrate values from raw PackStream representations into client objects.
        """
        graph = self.graph
        entities = self.entities
        keys = self.keys

        def hydrate_(obj, inst=None):
            if isinstance(obj, Structure):
                tag = obj.tag
                fields = obj.fields
                if tag == b"N":
                    return hydrate_node(graph, fields[0], inst=inst,
                                        metadata={"labels": list(fields[1])}, data=hydrate_(fields[2]))
                elif tag == b"R":
                    return hydrate_relationship(graph, fields[0], inst=inst,
                                                start=fields[1], end=fields[2],
                                                type=fields[3], data=hydrate_(fields[4]))
                elif tag == b"P":
                    from py2neo.data import Path
                    nodes = [hydrate_(node) for node in fields[0]]
                    u_rels = [_unbound_relationship(*map(hydrate_, r)) for r in fields[1]]
                    sequence = fields[2]
                    last_node = nodes[0]
                    steps = [last_node]
                    for i, rel_index in enumerate(sequence[::2]):
                        next_node = nodes[sequence[2 * i + 1]]
                        if rel_index > 0:
                            u_rel = u_rels[rel_index - 1]
                            rel = hydrate_relationship(graph, u_rel.id,
                                                       start=last_node.identity, end=next_node.identity,
                                                       type=u_rel.type, data=u_rel.properties)
                        else:
                            u_rel = u_rels[-rel_index - 1]
                            rel = hydrate_relationship(graph, u_rel.id,
                                                       start=next_node.identity, end=last_node.identity,
                                                       type=u_rel.type, data=u_rel.properties)
                        steps.append(rel)
                        steps.append(next_node)
                        last_node = next_node
                    return Path(*steps)
                else:
                    # Defer everything else to the official driver
                    return super(PackStreamHydrator, self).hydrate([obj])[0]
            elif isinstance(obj, list):
                return list(map(hydrate_, obj))
            elif isinstance(obj, dict):
                return {key: hydrate_(value) for key, value in obj.items()}
            else:
                return obj

        return tuple(hydrate_(value, entities.get(keys[i])) for i, value in enumerate(values))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

try:
    from configparser import SafeConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

try:
    from urllib.parse import urlparse, urlsplit
except ImportError:
    from urlparse import urlparse, urlsplit

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

from io import StringIO
import os
from sys import version_info


if version_info >= (3,):
    # Python 3

    atomic_types = (bool, bytearray, bytes, float, int, str)
    bytes_types = (bytearray, bytes)
    integer_types = (int,)
    list_types = (list, map)
    numeric_types = (int, float)
    string_types = (bytes, str)
    unicode_types = (str,)
    utf8_types = ()

    long = int
    uchr = chr

    def bstr(s, encoding="utf-8"):
        """ Convert a value to a byte string, held in a Python `bytearray` object.
        """
        if isinstance(s, bytearray):
            return s
        elif isinstance(s, bytes):
            return bytearray(s)
        elif isinstance(s, str):
            return bytearray(s.encode(encoding=encoding))
        else:
            try:
                return bytearray(s.__bytes__())
            except AttributeError:
                return bytearray(str(s).encode(encoding=encoding))

    def ustr(s, encoding="utf-8"):
        """ Convert a value to a Unicode string, held in a Python `str` object.
        """
        if isinstance(s, str):
            return s
        elif isinstance(s, (bytes, bytearray)):
            return s.decode(encoding=encoding)
        else:
            try:
                return s.__str__()
            except AttributeError:
                return str(s, encoding=encoding)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            return s.decode(encoding)
        else:
            return str(s)

    class PropertiesParser(SafeConfigParser):

        def read_properties(self, filename, section=None):
            if not section:
                basename = os.path.basename(filename)
                if basename.endswith(".properties"):
                    section = basename[:-11]
                else:
                    section = basename
            with open(filename) as f:
                data = f.read()
            self.read_string("[%s]\n%s" % (section, data), filename)

else:
    # Python 2

    atomic_types = (bool, bytearray, float, int, long, str, unicode)
    bytes_types = (bytearray,)
    integer_types = (int, long)
    list_types = (list,)
    numeric_types = (int, long, float)
    string_types = (str, unicode)
    unicode_types = (unicode,)
    utf8_types = (str,)

    long = long
    uchr = unichr

    def bstr(s, encoding="utf-8"):
        """ Convert a value to byte string, held in a Python `bytearray` object.
        """
        if isinstance(s, bytearray):
            return s
        elif isinstance(s, bytes):
            return bytearray(s)
        elif isinstance(s, unicode):
            return bytearray(s.encode(encoding=encoding))
        else:
            try:
                return bytearray(s.__bytes__())
            except AttributeError:
                return bytearray(unicode(s).encode(encoding=encoding))

    def ustr(s, encoding="utf-8"):
        """ Convert a value to a Unicode string, held in a Python `unicode` object.
        """
        if isinstance(s, unicode):
            return s
        elif isinstance(s, (bytes, bytearray)):
            return s.decode(encoding=encoding)
        else:
            try:
                return s.__unicode__()
            except AttributeError:
                return str(s).decode(encoding=encoding)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        if isinstance(s, str):
            return s
        else:
            return unicode(s).encode(encoding)

    class PropertiesParser(SafeConfigParser):

        def read_properties(self, filename, section=None):
            if not section:
                basename = os.path.basename(filename)
                if basename.endswith(".properties"):
                    section = basename[:-11]
                else:
                    section = basename
            data = StringIO()
            data.write("[%s]\n" % section)
            with codecs.open(filename, encoding="utf-8") as f:
                data.write(f.read())
            data.seek(0, os.SEEK_SET)
            self.readfp(data)



#!/usr/bin/env python
# coding: utf-8

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from re import compile as re_compile


version_string_pattern = re_compile("(\d+)\.(\d+)\.(\d+)-?(.*)")


class Version(tuple):

    def __new__(cls, string):
        return tuple.__new__(cls, version_string_pattern.match(string).groups())

    def __str__(self):
        return ".".join(self[:3]) + "".join("-%s" % part for part in self[3:] if part)

    def __eq__(self, other):
        return (int(self[0]), int(self[1]), int(self[2]), self[3]) == (int(other[0]), int(other[1]), int(other[2]), other[3])

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return (int(self[0]), int(self[1]), int(self[2]), self[3]) < (int(other[0]), int(other[1]), int(other[2]), other[3])

    @property
    def major(self):
        return int(self[0])

    @property
    def minor(self):
        return int(self[1])

    @property
    def patch(self):
        return int(self[2])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ["__author__", "__copyright__", "__email__", "__license__", "__package__", "__version__",
           "bolt_user_agent", "http_user_agent"]

__author__ = "Nigel Small <technige@nige.tech>"
__copyright__ = "2011-2018, Nigel Small"
__email__ = "py2neo@nige.tech"
__license__ = "Apache License, Version 2.0"
__package__ = "py2neo"
__version__ = "4.1.2"


def bolt_user_agent():
    from sys import platform, version_info
    import neo4j
    return "{}/{} neo4j-python/{} Python/{}.{}.{}-{}-{} ({})".format(
        *((__package__, __version__, neo4j.__version__,) + tuple(version_info) + (platform,)))


def http_user_agent():
    from sys import platform, version_info
    import urllib3
    return "{}/{} urllib3/{} Python/{}.{}.{}-{}-{} ({})".format(
        *((__package__, __version__, urllib3.__version__,) + tuple(version_info) + (platform,)))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from collections import deque
from datetime import datetime
from time import sleep
from warnings import warn

from py2neo.cypher import cypher_escape
from py2neo.data import Table, Record
from py2neo.internal.addressing import get_connection_data
from py2neo.internal.caching import ThreadLocalEntityCache
from py2neo.internal.compat import string_types, xstr
from py2neo.internal.util import version_tuple, title_case, snake_case
from py2neo.matching import NodeMatcher, RelationshipMatcher


update_stats_keys = [
    "constraints_added",
    "constraints_removed",
    "indexes_added",
    "indexes_removed",
    "labels_added",
    "labels_removed",
    "nodes_created",
    "nodes_deleted",
    "properties_set",
    "relationships_deleted",
    "relationships_created",
]


class Database(object):
    """ Accessor for an entire Neo4j graph database installation over
    Bolt or HTTP. Within the py2neo object hierarchy, a :class:`.Database`
    contains a :class:`.Graph` in which most activity occurs. Currently,
    Neo4j only supports one `Graph` per `Database`.

    An explicit URI can be passed to the constructor::

        >>> from py2neo import Database
        >>> db = Database("bolt://camelot.example.com:7687")

    Alternatively, the default value of ``bolt://localhost:7687`` is
    used::

        >>> default_db = Database()
        >>> default_db
        <Database uri='bolt://localhost:7687'>

    """

    _instances = {}

    _driver = None
    _graphs = None

    @classmethod
    def forget_all(cls):
        """ Forget all cached :class:`.Database` details.
        """
        for _, db in cls._instances.items():
            db._driver.close()
            db._driver = None
        cls._instances.clear()

    def __new__(cls, uri=None, **settings):
        connection_data = get_connection_data(uri, **settings)
        key = connection_data["hash"]
        try:
            inst = cls._instances[key]
        except KeyError:
            inst = super(Database, cls).__new__(cls)
            inst._connection_data = connection_data
            from py2neo.internal.http import HTTPDriver, HTTPSDriver
            from neo4j.v1 import Driver
            inst._driver = Driver(connection_data["uri"],
                                  auth=connection_data["auth"],
                                  encrypted=connection_data["secure"],
                                  user_agent=connection_data["user_agent"])
            inst._graphs = {}
            cls._instances[key] = inst
        return inst

    def __repr__(self):
        class_name = self.__class__.__name__
        data = self._connection_data
        return "<%s uri=%r secure=%r user_agent=%r>" % (
            class_name, data["uri"], data["secure"], data["user_agent"])

    def __eq__(self, other):
        try:
            return self.uri == other.uri
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._connection_data["hash"])

    def __contains__(self, database):
        return database in self._graphs

    def __getitem__(self, database):
        if database == "data" and database not in self._graphs:
            self._graphs[database] = Graph(**self._connection_data)
        return self._graphs[database]

    def __setitem__(self, database, graph):
        self._graphs[database] = graph

    def __iter__(self):
        yield "data"

    @property
    def driver(self):
        return self._driver

    @property
    def uri(self):
        """ The URI to which this `Database` is connected.
        """
        return self._connection_data["uri"]

    @property
    def default_graph(self):
        """ The default graph exposed by this database.

        :rtype: :class:`.Graph`
        """
        return self["data"]

    def keys(self):
        return list(self)

    def query_jmx(self, namespace, instance=None, name=None, type=None):
        """ Query the JMX service attached to this database.
        """
        d = {}
        for nom, _, attributes in self.default_graph.run("CALL dbms.queryJmx('')"):
            ns, _, terms = nom.partition(":")
            if ns != namespace:
                continue
            terms = dict(tuple(term.partition("=")[0::2]) for term in terms.split(","))
            if instance is not None and instance != terms["instance"]:
                continue
            if name is not None and name != terms["name"]:
                continue
            if type is not None and type != terms["type"]:
                continue
            for attr_name, attr_data in attributes.items():
                attr_value = attr_data.get("value")
                if attr_value == "true":
                    d[attr_name] = True
                elif attr_value == "false":
                    d[attr_name] = False
                elif isinstance(attr_value, string_types) and "." in attr_value:
                    try:
                        d[attr_name] = float(attr_value)
                    except (TypeError, ValueError):
                        d[attr_name] = attr_value
                else:
                    try:
                        d[attr_name] = int(attr_value)
                    except (TypeError, ValueError):
                        d[attr_name] = attr_value
        return d

    @property
    def name(self):
        """ Return the name of the active Neo4j database.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        return info.get("DatabaseName")

    @property
    def kernel_start_time(self):
        """ Return the time from which this Neo4j instance was in operational mode.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        return datetime.fromtimestamp(info["KernelStartTime"] / 1000.0)

    @property
    def kernel_version(self):
        """ Return the version of Neo4j.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        version_string = info["KernelVersion"].partition("version:")[-1].partition(",")[0].strip()
        return version_tuple(version_string)

    @property
    def product(self):
        """ Return the product name.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        return info["KernelVersion"]

    @property
    def store_creation_time(self):
        """ Return the time when this Neo4j graph store was created.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        return datetime.fromtimestamp(info["StoreCreationDate"] / 1000.0)

    @property
    def store_id(self):
        """ Return an identifier that, together with store creation time,
        uniquely identifies this Neo4j graph store.
        """
        info = self.query_jmx("org.neo4j", name="Kernel")
        return info["StoreId"]

    @property
    def primitive_counts(self):
        """ Return a dictionary of estimates of the numbers of different
        kinds of Neo4j primitives.
        """
        return self.query_jmx("org.neo4j", name="Primitive count")

    @property
    def store_file_sizes(self):
        """ Return a dictionary of file sizes for each file in the Neo4j
        graph store.
        """
        return self.query_jmx("org.neo4j", name="Store file sizes")

    @property
    def config(self):
        """ Return a dictionary of the configuration parameters used to
        configure Neo4j.
        """
        return self.query_jmx("org.neo4j", name="Configuration")


class Graph(object):
    """ The `Graph` class represents the graph data storage space within
    a Neo4j graph database. Connection details are provided using URIs
    and/or individual settings.

    Supported URI schemes are:

    - ``http``
    - ``https``
    - ``bolt``
    - ``bolt+routing``

    The full set of supported `settings` are:

    ==============  =============================================  ==============  =============
    Keyword         Description                                    Type            Default
    ==============  =============================================  ==============  =============
    ``auth``        A 2-tuple of (user, password)                  tuple           ``('neo4j', 'password')``
    ``host``        Database server host name                      str             ``'localhost'``
    ``password``    Password to use for authentication             str             ``'password'``
    ``port``        Database server port                           int             ``7687``
    ``scheme``      Use a specific URI scheme                      str             ``'bolt'``
    ``secure``      Use a secure connection (TLS)                  bool            ``False``
    ``user``        User to authenticate as                        str             ``'neo4j'``
    ``user_agent``  User agent to send for all connections         str             `(depends on URI scheme)`
    ==============  =============================================  ==============  =============

    Each setting can be provided as a keyword argument or as part of
    an ``http:``, ``https:``, ``bolt:`` or ``bolt+routing:`` URI. Therefore, the examples
    below are equivalent::

        >>> from py2neo import Graph
        >>> graph_1 = Graph()
        >>> graph_2 = Graph(host="localhost")
        >>> graph_3 = Graph("bolt://localhost:7687")

    Once obtained, the `Graph` instance provides direct or indirect
    access to most of the functionality available within py2neo.
    """

    #: The :class:`.Database` to which this :class:`.Graph` belongs.
    database = None

    #: The :class:`.Schema` resource for this :class:`.Graph`.
    schema = None

    node_cache = ThreadLocalEntityCache()
    relationship_cache = ThreadLocalEntityCache()

    def __new__(cls, uri=None, **settings):
        name = settings.pop("name", "data")
        database = Database(uri, **settings)
        if name in database:
            inst = database[name]
        else:
            inst = object.__new__(cls)
            inst.database = database
            inst.schema = Schema(inst)
            inst.__name__ = name
            database[name] = inst
        return inst

    def __repr__(self):
        return "<Graph database=%r name=%r>" % (self.database, self.__name__)

    def __eq__(self, other):
        try:
            return self.database == other.database and self.__name__ == other.__name__
        except (AttributeError, TypeError):
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.relationships)

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def begin(self, autocommit=False):
        """ Begin a new :class:`.Transaction`.

        :param autocommit: if :py:const:`True`, the transaction will
                         automatically commit after the first operation
        """
        return Transaction(self, autocommit)

    def create(self, subgraph):
        """ Run a :meth:`.Transaction.create` operation within a
        :class:`.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        with self.begin() as tx:
            tx.create(subgraph)

    def delete(self, subgraph):
        """ Run a :meth:`.Transaction.delete` operation within an
        `autocommit` :class:`.Transaction`. To delete only the
        relationships, use the :meth:`.separate` method.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        """
        self.begin(autocommit=True).delete(subgraph)

    def delete_all(self):
        """ Delete all nodes and relationships from this :class:`.Graph`.

        .. warning::
            This method will permanently remove **all** nodes and relationships
            from the graph and cannot be undone.
        """
        self.run("MATCH (a) DETACH DELETE a")
        self.node_cache.clear()
        self.relationship_cache.clear()

    def evaluate(self, cypher, parameters=None, **kwparameters):
        """ Run a :meth:`.Transaction.evaluate` operation within an
        `autocommit` :class:`.Transaction`.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :return: first value from the first record returned or
                 :py:const:`None`.
        """
        return self.begin(autocommit=True).evaluate(cypher, parameters, **kwparameters)

    def exists(self, subgraph):
        """ Run a :meth:`.Transaction.exists` operation within an
        `autocommit` :class:`.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :return:
        """
        return self.begin(autocommit=True).exists(subgraph)

    def match(self, nodes=None, r_type=None, limit=None):
        """ Match and return all relationships with specific criteria.

        For example, to find all of Alice's friends::

            for rel in graph.match((alice, ), r_type="FRIEND"):
                print(rel.end_node["name"])

        :param nodes: Sequence or Set of start and end nodes (:const:`None` means any node);
                a Set implies a match in any direction
        :param r_type: type of relationships to match (:const:`None` means any type)
        :param limit: maximum number of relationships to match (:const:`None` means unlimited)
        """
        return RelationshipMatcher(self).match(nodes=nodes, r_type=r_type).limit(limit)

    def match_one(self, nodes=None, r_type=None):
        """ Match and return one relationship with specific criteria.

        :param nodes: Sequence or Set of start and end nodes (:const:`None` means any node);
                a Set implies a match in any direction
        :param r_type: type of relationships to match (:const:`None` means any type)
        """
        rels = list(self.match(nodes=nodes, r_type=r_type, limit=1))
        if rels:
            return rels[0]
        else:
            return None

    def merge(self, subgraph, label=None, *property_keys):
        """ Run a :meth:`.Transaction.merge` operation within an
        `autocommit` :class:`.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :param label: label on which to match any existing nodes
        :param property_keys: property keys on which to match any existing nodes
        """
        with self.begin() as tx:
            tx.merge(subgraph, label, *property_keys)

    @property
    def name(self):
        return self.__name__

    @property
    def nodes(self):
        """ Obtain a :class:`.NodeMatcher` for this graph.

        This can be used to find nodes that match given criteria:

            >>> graph = Graph()
            >>> graph.nodes[1234]
            (_1234:Person {name: 'Alice'})
            >>> graph.nodes.get(1234)
            (_1234:Person {name: 'Alice'})
            >>> graph.nodes.match("Person", name="Alice").first()
            (_1234:Person {name: 'Alice'})

        Nodes can also be efficiently counted using this attribute:

            >>> len(graph.nodes)
            55691
            >>> len(graph.nodes.match("Person", age=33))
            12

        """
        return NodeMatcher(self)

    def pull(self, subgraph):
        """ Pull data to one or more entities from their remote counterparts.

        :param subgraph: the collection of nodes and relationships to pull
        """
        with self.begin() as tx:
            tx.pull(subgraph)

    def push(self, subgraph):
        """ Push data from one or more entities to their remote counterparts.

        :param subgraph: the collection of nodes and relationships to push
        """
        with self.begin() as tx:
            tx.push(subgraph)

    @property
    def relationships(self):
        """ Obtain a :class:`.RelationshipMatcher` for this graph.

        This can be used to find relationships that match given criteria
        as well as efficiently count relationships.
        """
        return RelationshipMatcher(self)

    def run(self, cypher, parameters=None, **kwparameters):
        """ Run a :meth:`.Transaction.run` operation within an
        `autocommit` :class:`.Transaction`.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :param kwparameters: extra keyword parameters
        :return:
        """
        return self.begin(autocommit=True).run(cypher, parameters, **kwparameters)

    def separate(self, subgraph):
        """ Run a :meth:`.Transaction.separate` operation within an
        `autocommit` :class:`.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        self.begin(autocommit=True).separate(subgraph)


class Schema(object):
    """ The schema resource attached to a `Graph` instance.
    """

    def __init__(self, graph):
        self.graph = graph

    @property
    def node_labels(self):
        """ The set of node labels currently defined within the graph.
        """
        return frozenset(record[0] for record in self.graph.run("CALL db.labels"))

    @property
    def relationship_types(self):
        """ The set of relationship types currently defined within the graph.
        """
        return frozenset(record[0] for record in self.graph.run("CALL db.relationshipTypes"))

    def create_index(self, label, *property_keys):
        """ Create a schema index for a label and property
        key combination.
        """
        self.graph.run("CREATE INDEX ON :%s(%s)" %
                       (cypher_escape(label), ",".join(map(cypher_escape, property_keys)))).close()
        while property_keys not in self.get_indexes(label):
            sleep(0.1)

    def create_uniqueness_constraint(self, label, *property_keys):
        """ Create a uniqueness constraint for a label.
        """
        self.graph.run("CREATE CONSTRAINT ON (a:%s) "
                       "ASSERT a.%s IS UNIQUE" %
                       (cypher_escape(label), ",".join(map(cypher_escape, property_keys)))).close()
        while property_keys not in self.get_uniqueness_constraints(label):
            sleep(0.1)

    def drop_index(self, label, *property_keys):
        """ Remove label index for a given property key.
        """
        self.graph.run("DROP INDEX ON :%s(%s)" %
                       (cypher_escape(label), ",".join(map(cypher_escape, property_keys)))).close()

    def drop_uniqueness_constraint(self, label, *property_keys):
        """ Remove the uniqueness constraint for a given property key.
        """
        self.graph.run("DROP CONSTRAINT ON (a:%s) "
                       "ASSERT a.%s IS UNIQUE" %
                       (cypher_escape(label), ",".join(map(cypher_escape, property_keys)))).close()

    def _get_indexes(self, label, t=None):
        indexes = []
        for record in self.graph.run("CALL db.indexes"):
            lbl = None
            properties = []
            if len(record) == 6:
                description, lbl, properties, state, typ, provider = record
            elif len(record) == 3:
                description, state, typ = record
            else:
                raise RuntimeError("Unexpected response from procedure db.indexes (%d fields)" % len(record))
            if state not in (u"ONLINE", u"online"):
                continue
            if t and typ != t:
                continue
            if not lbl or not properties:
                from py2neo.cypher.lexer import CypherLexer
                from pygments.token import Token
                tokens = list(CypherLexer().get_tokens(description))
                for token_type, token_value in tokens:
                    if token_type is Token.Name.Label:
                        lbl = token_value.strip("`")
                    elif token_type is Token.Name.Variable:
                        properties.append(token_value.strip("`"))
            if not lbl or not properties:
                continue
            if lbl == label:
                indexes.append(tuple(properties))
        return indexes

    def get_indexes(self, label):
        """ Fetch a list of indexed property keys for a label.
        """
        return self._get_indexes(label)

    def get_uniqueness_constraints(self, label):
        """ Fetch a list of unique constraints for a label.
        """
        return self._get_indexes(label, "node_unique_property")


class Result(object):
    """ Wraps a BoltStatementResult
    """

    def __init__(self, graph, entities, result):
        from neo4j.v1 import BoltStatementResult
        from py2neo.internal.http import HTTPStatementResult
        from py2neo.internal.packstream import PackStreamHydrator
        self.result = result
        self.result.error_class = GraphError.hydrate
        # TODO: un-yuk this
        if isinstance(result, HTTPStatementResult):
            self.result._hydrant.entities = entities
            self.result_iterator = iter(self.result)
        elif isinstance(result, BoltStatementResult):
            self.result._hydrant = PackStreamHydrator(graph, result.keys(), entities)
            self.result_iterator = iter(map(Record, self.result))
        else:
            raise RuntimeError("Unexpected statement result class %r" % result.__class__.__name__)

    def keys(self):
        """ Return the keys for the whole data set.
        """
        return self.result.keys()

    def summary(self):
        """ Return the summary.
        """
        return self.result.summary()

    def plan(self):
        """ Return the query plan, if available.
        """
        metadata = self.result.summary().metadata
        plan = {}
        if "plan" in metadata:
            plan.update(metadata["plan"])
        if "profile" in metadata:
            plan.update(metadata["profile"])
        if "http_plan" in metadata:
            plan.update(metadata["http_plan"]["root"])

        def collapse_args(data):
            if "args" in data:
                for key in data["args"]:
                    data[key] = data["args"][key]
                del data["args"]
            if "children" in data:
                for child in data["children"]:
                    collapse_args(child)

        def snake_keys(data):
            if isinstance(data, list):
                for item in data:
                    snake_keys(item)
                return
            if not isinstance(data, dict):
                return
            for key, value in list(data.items()):
                new_key = snake_case(key)
                if new_key != key:
                    data[new_key] = value
                    del data[key]
                if isinstance(value, (list, dict)):
                    snake_keys(value)

        collapse_args(plan)
        snake_keys(plan)
        return plan

    def stats(self):
        """ Return the query statistics.
        """
        return vars(self.result.summary().counters)

    def fetch(self):
        """ Fetch and return the next item.
        """
        try:
            return next(self.result_iterator)
        except StopIteration:
            return None


class GraphError(Exception):
    """
    """

    __cause__ = None

    http_status_code = None
    code = None
    message = None

    @classmethod
    def hydrate(cls, data):
        code = data["code"]
        message = data["message"]
        _, classification, category, title = code.split(".")
        if classification == "ClientError":
            try:
                error_cls = ClientError.get_mapped_class(code)
            except KeyError:
                error_cls = ClientError
                message = "%s: %s" % (title_case(title), message)
        elif classification == "DatabaseError":
            error_cls = DatabaseError
        elif classification == "TransientError":
            error_cls = TransientError
        else:
            error_cls = cls
        inst = error_cls(message)
        inst.code = code
        inst.message = message
        return inst

    def __new__(cls, *args, **kwargs):
        try:
            exception = kwargs["exception"]
            error_cls = type(xstr(exception), (cls,), {})
        except KeyError:
            error_cls = cls
        return Exception.__new__(error_cls, *args)

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        for key, value in kwargs.items():
            setattr(self, key.lower(), value)


class ClientError(GraphError):
    """ The Client sent a bad request - changing the request might yield a successful outcome.
    """

    @classmethod
    def get_mapped_class(cls, status):
        from neo4j.exceptions import ConstraintError, CypherSyntaxError, CypherTypeError, Forbidden, AuthError
        return {

            # ConstraintError
            "Neo.ClientError.Schema.ConstraintValidationFailed": ConstraintError,
            "Neo.ClientError.Schema.ConstraintViolation": ConstraintError,
            "Neo.ClientError.Statement.ConstraintVerificationFailed": ConstraintError,
            "Neo.ClientError.Statement.ConstraintViolation": ConstraintError,

            # CypherSyntaxError
            "Neo.ClientError.Statement.InvalidSyntax": CypherSyntaxError,
            "Neo.ClientError.Statement.SyntaxError": CypherSyntaxError,

            # CypherTypeError
            "Neo.ClientError.Procedure.TypeError": CypherTypeError,
            "Neo.ClientError.Statement.InvalidType": CypherTypeError,
            "Neo.ClientError.Statement.TypeError": CypherTypeError,

            # Forbidden
            "Neo.ClientError.General.ForbiddenOnReadOnlyDatabase": Forbidden,
            "Neo.ClientError.General.ReadOnly": Forbidden,
            "Neo.ClientError.Schema.ForbiddenOnConstraintIndex": Forbidden,
            "Neo.ClientError.Schema.IndexBelongsToConstrain": Forbidden,
            "Neo.ClientError.Security.Forbidden": Forbidden,
            "Neo.ClientError.Transaction.ForbiddenDueToTransactionType": Forbidden,

            # Unauthorized
            "Neo.ClientError.Security.AuthorizationFailed": AuthError,
            "Neo.ClientError.Security.Unauthorized": AuthError,

        }[status]


class DatabaseError(GraphError):
    """ The database failed to service the request.
    """


class TransientError(GraphError):
    """ The database cannot service the request right now, retrying later might yield a successful outcome.
    """


class TransactionFinished(GraphError):
    """ Raised when actions are attempted against a :class:`.Transaction`
    that is no longer available for use.
    """


class Transaction(object):
    """ A transaction is a logical container for multiple Cypher statements.
    """

    session = None

    _finished = False

    def __init__(self, graph, autocommit=False):
        self.graph = graph
        self.autocommit = autocommit
        self.entities = deque()
        self.driver = driver = self.graph.database.driver
        self.session = driver.session()
        self.results = []
        if autocommit:
            self.transaction = None
        else:
            self.transaction = self.session.begin_transaction()

    def __del__(self):
        if self.session:
            self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

    def _assert_unfinished(self):
        if self._finished:
            raise TransactionFinished(self)

    def finished(self):
        """ Indicates whether or not this transaction has been completed
        or is still open.
        """
        return self._finished

    def run(self, cypher, parameters=None, **kwparameters):
        """ Send a Cypher statement to the server for execution and return
        a :py:class:`.Cursor` for navigating its result.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :returns: :py:class:`.Cursor` object
        """
        from neo4j.v1 import CypherError

        self._assert_unfinished()
        try:
            entities = self.entities.popleft()
        except IndexError:
            entities = {}

        try:
            if self.transaction:
                result = self.transaction.run(cypher, parameters, **kwparameters)
            else:
                result = self.session.run(cypher, parameters, **kwparameters)
        except CypherError as error:
            raise GraphError.hydrate({"code": error.code, "message": error.message})
        else:
            r = Result(self.graph, entities, result)
            self.results.append(r)
            return Cursor(r)
        finally:
            if not self.transaction:
                self.finish()

    def process(self):
        """ Send all pending statements to the server for processing.
        """
        self._assert_unfinished()
        self.session.sync()

    def finish(self):
        self.process()
        if self.transaction:
            self.transaction.close()
        self._assert_unfinished()
        self._finished = True
        self.session.close()
        self.session = None

    def commit(self):
        """ Commit the transaction.
        """
        if self.transaction:
            self.transaction.success = True
        self.finish()

    def rollback(self):
        """ Roll back the current transaction, undoing all actions previously taken.
        """
        self._assert_unfinished()
        if self.transaction:
            self.transaction.success = False
        self.finish()

    def evaluate(self, cypher, parameters=None, **kwparameters):
        """ Execute a single Cypher statement and return the value from
        the first column of the first record.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :returns: single return value or :const:`None`
        """
        return self.run(cypher, parameters, **kwparameters).evaluate(0)

    def create(self, subgraph):
        """ Create remote nodes and relationships that correspond to those in a
        local subgraph. Any entities in *subgraph* that are already bound to
        remote entities will remain unchanged, those which are not will become
        bound to their newly-created counterparts.

        For example::

            >>> from py2neo import Graph, Node, Relationship
            >>> g = Graph()
            >>> tx = g.begin()
            >>> a = Node("Person", name="Alice")
            >>> tx.create(a)
            >>> b = Node("Person", name="Bob")
            >>> ab = Relationship(a, "KNOWS", b)
            >>> tx.create(ab)
            >>> tx.commit()
            >>> g.exists(ab)
            True

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                    creatable object
        """
        try:
            create = subgraph.__db_create__
        except AttributeError:
            raise TypeError("No method defined to create object %r" % subgraph)
        else:
            create(self)

    def delete(self, subgraph):
        """ Delete the remote nodes and relationships that correspond to
        those in a local subgraph. To delete only the relationships, use
        the :meth:`.separate` method.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            delete = subgraph.__db_delete__
        except AttributeError:
            raise TypeError("No method defined to delete object %r" % subgraph)
        else:
            delete(self)

    def exists(self, subgraph):
        """ Determine whether one or more graph entities all exist within the
        database. Note that if any nodes or relationships in *subgraph* are not
        bound to remote counterparts, this method will return ``False``.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        :returns: ``True`` if all entities exist remotely, ``False`` otherwise
        """
        try:
            exists = subgraph.__db_exists__
        except AttributeError:
            raise TypeError("No method defined to check existence of object %r" % subgraph)
        else:
            return exists(self)

    def merge(self, subgraph, primary_label=None, primary_key=None):
        """ Merge nodes and relationships from a local subgraph into the
        database. Each node and relationship is merged independently, with
        nodes merged first and relationships merged second.

        For each node, the merge is carried out by comparing that node with a
        potential remote equivalent on the basis of a label and property value.
        If no remote match is found, a new node is created. The label and
        property to use for comparison are determined by `primary_label` and
        `primary_key` but may be overridden for individual nodes by the
        presence of `__primarylabel__` and `__primarykey__` attributes on
        the node itself.

        For each relationship, the merge is carried out by comparing that
        relationship with a potential remote equivalent on the basis of matching
        start and end nodes plus relationship type. If no remote match is found,
        a new relationship is created.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :param primary_label: label on which to match any existing nodes
        :param primary_key: property key(s) on which to match any existing
                            nodes
        """
        try:
            merge = subgraph.__db_merge__
        except AttributeError:
            raise TypeError("No method defined to merge object %r" % subgraph)
        else:
            merge(self, primary_label, primary_key)

    def pull(self, subgraph):
        """ Update local entities from their remote counterparts.

        For any nodes and relationships that exist in both the local
        :class:`.Subgraph` and the remote :class:`.Graph`, pull properties
        and node labels into the local copies. This operation does not
        create or delete any entities.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            pull = subgraph.__db_pull__
        except AttributeError:
            raise TypeError("No method defined to pull object %r" % subgraph)
        else:
            return pull(self)

    def push(self, subgraph):
        """ Update remote entities from their local counterparts.

        For any nodes and relationships that exist in both the local
        :class:`.Subgraph` and the remote :class:`.Graph`, push properties
        and node labels into the remote copies. This operation does not
        create or delete any entities.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            push = subgraph.__db_push__
        except AttributeError:
            raise TypeError("No method defined to push object %r" % subgraph)
        else:
            return push(self)

    def separate(self, subgraph):
        """ Delete the remote relationships that correspond to those in a local
        subgraph. This leaves any nodes untouched.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            separate = subgraph.__db_separate__
        except AttributeError:
            raise TypeError("No method defined to separate object %r" % subgraph)
        else:
            separate(self)


class Cursor(object):
    """ A `Cursor` is a navigator for a stream of records.

    A cursor can be thought of as a window onto an underlying data
    stream. All cursors in py2neo are "forward-only", meaning that
    navigation starts before the first record and may proceed only in a
    forward direction.

    It is not generally necessary for application code to instantiate a
    cursor directly as one will be returned by any Cypher execution method.
    However, cursor creation requires only a :class:`.DataSource` object
    which contains the logic for how to access the source data that the
    cursor navigates.

    Many simple cursor use cases require only the :meth:`.forward` method
    and the :attr:`.current` attribute. To navigate through all available
    records, a `while` loop can be used::

        while cursor.forward():
            print(cursor.current["name"])

    If only the first record is of interest, a similar `if` structure will
    do the job::

        if cursor.forward():
            print(cursor.current["name"])

    To combine `forward` and `current` into a single step, use the built-in
    py:func:`next` function::

        print(next(cursor)["name"])

    Cursors are also iterable, so can be used in a loop::

        for record in cursor:
            print(record["name"])

    For queries that are expected to return only a single value within a
    single record, use the :meth:`.evaluate` method. This will return the
    first value from the next record or :py:const:`None` if neither the
    field nor the record are present::

        print(cursor.evaluate())

    """

    def __init__(self, result):
        self._result = result
        self._current = None

    def __next__(self):
        if self.forward():
            return self._current
        else:
            raise StopIteration()

    # Exists only for Python 2 iteration compatibility
    next = __next__

    def __iter__(self):
        while self.forward():
            yield self._current

    def __getitem__(self, key):
        return self._current[key]

    @property
    def current(self):
        """ Returns the current record or :py:const:`None` if no record
        has yet been selected.
        """
        return self._current

    def close(self):
        """ Close this cursor and free up all associated resources.
        """
        self._result = None
        self._current = None

    def keys(self):
        """ Return the field names for the records in the stream.
        """
        return self._result.keys()

    def summary(self):
        """ Return the result summary.
        """
        return self._result.summary()

    def plan(self):
        """ Return the plan returned with this result, if any.
        """
        return self._result.plan()

    def stats(self):
        """ Return the query statistics.
        """
        s = dict.fromkeys(update_stats_keys, 0)
        s.update(self._result.stats())
        s["contains_updates"] = bool(sum(s.get(k, 0) for k in update_stats_keys))
        return s

    def forward(self, amount=1):
        """ Attempt to move the cursor one position forward (or by
        another amount if explicitly specified). The cursor will move
        position by up to, but never more than, the amount specified.
        If not enough scope for movement remains, only that remainder
        will be consumed. The total amount moved is returned.

        :param amount: the amount to move the cursor
        :returns: the amount that the cursor was able to move
        """
        if amount == 0:
            return 0
        assert amount > 0
        amount = int(amount)
        moved = 0
        fetch = self._result.fetch
        while moved != amount:
            new_current = fetch()
            if new_current is None:
                break
            else:
                self._current = new_current
                moved += 1
        return moved

    def evaluate(self, field=0):
        """ Return the value of the first field from the next record
        (or the value of another field if explicitly specified).

        This method attempts to move the cursor one step forward and,
        if successful, selects and returns an individual value from
        the new current record. By default, this value will be taken
        from the first value in that record but this can be overridden
        with the `field` argument, which can represent either a
        positional index or a textual key.

        If the cursor cannot be moved forward or if the record contains
        no values, :py:const:`None` will be returned instead.

        This method is particularly useful when it is known that a
        Cypher query returns only a single value.

        :param field: field to select value from (optional)
        :returns: value of the field or :py:const:`None`

        Example:
            >>> from py2neo import Graph
            >>> g = Graph()
            >>> g.run("MATCH (a) WHERE a.email={x} RETURN a.name", x="bob@acme.com").evaluate()
            'Bob Robertson'
        """
        if self.forward():
            try:
                return self[field]
            except IndexError:
                return None
        else:
            return None

    def data(self):
        """ Consume and extract the entire result as a list of
        dictionaries.

        ::

            >>> from py2neo import Graph
            >>> graph = Graph()
            >>> graph.run("MATCH (a:Person) RETURN a.name, a.born LIMIT 4").data()
            [{'a.born': 1964, 'a.name': 'Keanu Reeves'},
             {'a.born': 1967, 'a.name': 'Carrie-Anne Moss'},
             {'a.born': 1961, 'a.name': 'Laurence Fishburne'},
             {'a.born': 1960, 'a.name': 'Hugo Weaving'}]

        :return: the full query result
        :rtype: `list` of `dict`
        """
        return [record.data() for record in self]

    def to_table(self):
        """ Consume and extract the entire result as a :class:`.Table`
        object.

        :return: the full query result
        """
        return Table(self)

    def to_subgraph(self):
        """ Consume and extract the entire result as a :class:`.Subgraph`
        containing the union of all the graph structures within.

        :return: :class:`.Subgraph` object
        """
        s = None
        for record in self:
            s_ = record.to_subgraph()
            if s_ is not None:
                if s is None:
                    s = s_
                else:
                    s |= s_
        return s

    def to_ndarray(self, dtype=None, order='K'):
        """ Consume and extract the entire result as a
        `numpy.ndarray <https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html>`_.

        .. note::
           This method requires `numpy` to be installed.

        :param dtype:
        :param order:
        :warns: If `numpy` is not installed
        :returns: `ndarray <https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html>`__ object.
        """
        try:
            from numpy import array
        except ImportError:
            warn("Numpy is not installed.")
            raise
        else:
            return array(list(map(list, self)), dtype=dtype, order=order)

    def to_series(self, field=0, index=None, dtype=None):
        """ Consume and extract one field of the entire result as a
        `pandas.Series <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`_.

        .. note::
           This method requires `pandas` to be installed.

        :param field:
        :param index:
        :param dtype:
        :warns: If `pandas` is not installed
        :returns: `Series <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
        """
        try:
            from pandas import Series
        except ImportError:
            warn("Pandas is not installed.")
            raise
        else:
            return Series([record[field] for record in self], index=index, dtype=dtype)

    def to_data_frame(self, index=None, columns=None, dtype=None):
        """ Consume and extract the entire result as a
        `pandas.DataFrame <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#dataframe>`_.

        ::

            >>> from py2neo import Graph
            >>> graph = Graph()
            >>> graph.run("MATCH (a:Person) RETURN a.name, a.born LIMIT 4").to_data_frame()
               a.born              a.name
            0    1964        Keanu Reeves
            1    1967    Carrie-Anne Moss
            2    1961  Laurence Fishburne
            3    1960        Hugo Weaving

        .. note::
           This method requires `pandas` to be installed.

        :param index: Index to use for resulting frame.
        :param columns: Column labels to use for resulting frame.
        :param dtype: Data type to force.
        :warns: If `pandas` is not installed
        :returns: `DataFrame <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
        """
        try:
            from pandas import DataFrame
        except ImportError:
            warn("Pandas is not installed.")
            raise
        else:
            return DataFrame(list(map(dict, self)), index=index, columns=columns, dtype=dtype)

    def to_matrix(self, mutable=False):
        """ Consume and extract the entire result as a
        `sympy.Matrix <http://docs.sympy.org/latest/tutorial/matrices.html>`_.

        .. note::
           This method requires `sympy` to be installed.

        :param mutable:
        :returns: `Matrix <http://docs.sympy.org/latest/tutorial/matrices.html>`_ object.
        """
        try:
            from sympy import MutableMatrix, ImmutableMatrix
        except ImportError:
            warn("Sympy is not installed.")
            raise
        else:
            if mutable:
                return MutableMatrix(list(map(list, self)))
            else:
                return ImmutableMatrix(list(map(list, self)))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from os.path import dirname, join as path_join
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

from py2neo.meta import __author__, __email__, __license__, __package__, __version__


with open(path_join(dirname(__file__), "README.rst")) as f:
    README = f.read()

packages = find_packages(exclude=("docs", "test"))
package_metadata = {
    "name": "py2neo-archive",
    "version": __version__,
    "description": "Python client library and toolkit for Neo4j",
    "long_description": README,
    "long_description_content_type": "text/markdown",
    "author": __author__,
    "author_email": __email__,
    "url": "http://py2neo.org/",
    "entry_points": {
        "console_scripts": [
            "py2neo = py2neo.__main__:main",
        ],
        "pygments.lexers": [
            "py2neo.cypher = py2neo.cypher.lexer:CypherLexer",
        ],
    },
    "packages": packages,
    "py_modules": [],
    "install_requires": [
        "click>=2.0",
        "colorama",
        "neo4j-driver>=1.6.2,<1.7",
        "neotime==1.0.0",
        "prompt_toolkit==1.0.15",
        "pygments>=2.0",
        "pytest",
        "urllib3[secure]<1.23,>=1.21.1",
    ],
    "extras_require": {
    },
    "license": __license__,
    "classifiers": [
        "Development Status :: 6 - Mature",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
    ],
    "zip_safe": False,
}

setup(**package_metadata)



#!/usr/bin/env python
# coding: utf-8

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from hashlib import sha256
from os import curdir, getenv, kill, listdir, makedirs, rename
from os.path import abspath, dirname, expanduser, isdir, isfile, join as path_join
from random import randint
from re import compile as re_compile
from shutil import rmtree
from socket import create_connection
from subprocess import check_call, check_output, CalledProcessError
from tarfile import TarFile, ReadError
from time import sleep, time
from warnings import warn

from py2neo.admin.dist import Distribution, archive_format
from py2neo.internal.compat import bstr
from py2neo.internal.util import hex_bytes, unhex_bytes


class Warehouse(object):
    """ A local storage area for Neo4j installations.
    """

    def __init__(self, home=None):
        self.home = home or getenv("PY2NEO_HOME") or expanduser("~/.py2neo")
        self.dist = path_join(self.home, "dist")
        self.run = path_join(self.home, "run")
        self.cc = path_join(self.home, "cc")

    def get(self, name, database=None, role=None, member=None):
        """ Obtain a Neo4j installation by name.

        :param name:
        :param database:
        :param role:
        :param member:
        :return:
        """
        if database and role and member is not None:
            container = path_join(self.cc, name, database, role, str(member))
        else:
            container = path_join(self.run, name)
        for dir_name in listdir(container):
            dir_path = path_join(container, dir_name)
            if isdir(dir_path):
                return Installation(dir_path)
        raise IOError("Could not locate installation directory")

    def install(self, name, edition=None, version=None, database=None, role=None, member=None):
        """ Install Neo4j.

        :param name:
        :param edition:
        :param version:
        :param database:
        :param role:
        :param member:
        :return:
        """
        if database and role and member is not None:
            container = path_join(self.cc, name, database, role, str(member))
        else:
            container = path_join(self.run, name)
        rmtree(container, ignore_errors=True)
        makedirs(container)
        archive_file = Distribution(edition, version).download(self.dist)
        try:
            with TarFile.open(archive_file, "r:{}".format(archive_format)) as archive:
                archive.extractall(container)
        except ReadError:
            # The tarfile module sometimes has trouble with certain tar
            # files for unknown reasons. This workaround falls back to
            # command line.
            check_call(["tar", "x", "-C", container, "-f", archive_file])
        return self.get(name, database, role, member)

    def uninstall(self, name, database=None, role=None, member=None):
        """ Remove a Neo4j installation.

        :param name:
        :param database:
        :param role:
        :param member:
        :return:
        """
        if database and role and member is not None:
            container = path_join(self.cc, name, database, role, str(member))
        else:
            container = path_join(self.run, name)
        rmtree(container, ignore_errors=True)

    def directory(self):
        """ Fetch a dictionary of :class:`.Installation` objects, keyed
        by name, for all available Neo4j installations.
        """
        try:
            return {name: self.get(name) for name in listdir(self.run) if not name.startswith(".")}
        except OSError:
            return {}

    def rename(self, name, new_name):
        """ Rename a Neo4j installation.

        :param name:
        :param new_name:
        :return:
        """
        rename(path_join(self.run, name), path_join(self.run, new_name))


class Installation(object):
    """ A Neo4j 3.0+ server installation.
    """

    config_file = "neo4j.conf"

    def __init__(self, home=None):
        self.home = home or abspath(curdir)
        self.server = Server(self)
        self.auth = AuthFile(path_join(self.home, "data", "dbms", "auth"))

    def __repr__(self):
        return "<%s home=%r>" % (self.__class__.__name__, self.home)

    @property
    def store_path(self):
        """ The location of the graph database store on disk.
        """
        return path_join(self.home, "data", "databases",
                         self.get_config("dbms.active_database", "graph.db"))

    def get_config(self, key, default=None):
        """ Retrieve the value of a configuration item.

        :param key:
        :param default:
        :return:
        """
        config_file_path = path_join(self.home, "conf", self.config_file)
        with open(config_file_path, "r") as f_in:
            for line in f_in:
                if line.startswith(key + "="):
                    return line.strip().partition("=")[-1]
        return default

    def set_config(self, key, value):
        """ Update a single configuration value.

        :param key:
        :param value:
        """
        self.update_config({key: value})

    def update_config(self, properties):
        """ Update multiple configuration values.

        :param properties:
        """
        config_file_path = path_join(self.home, "conf", self.config_file)
        with open(config_file_path, "r") as f_in:
            lines = f_in.readlines()
        with open(config_file_path, "w") as f_out:
            properties2 = dict(properties)
            for line in lines:
                for key, value in properties2.items():
                    if line.startswith(key + "=") or \
                            (line.startswith("#") and line[1:].lstrip().startswith(key + "=")):
                        if value is True:
                            value = "true"
                        if value is False:
                            value = "false"
                        f_out.write("%s=%s\n" % (key, value))
                        del properties2[key]
                        break
                else:
                    f_out.write(line)
            for key, value in properties2.items():
                if value is True:
                    value = "true"
                if value is False:
                    value = "false"
                f_out.write("%s=%s\n" % (key, value))

    @property
    def auth_enabled(self):
        """ Settable boolean property for enabling and disabling auth
        on this server.
        """
        return self.get_config("dbms.security.auth_enabled", "true") == "true"

    @auth_enabled.setter
    def auth_enabled(self, value):
        self.set_config("dbms.security.auth_enabled", value)

    def _get_protocol_address(self, protocol, default_port):
        if self.get_config("dbms.connector.%s.enabled" % protocol, "true") != "true":
            raise ValueError("Protocol %r not enabled" % protocol)
        address = self.get_config("dbms.connector.%s.listen_address" % protocol)
        if address:
            host, _, port = address.partition(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return host or "localhost", port
        return "localhost", default_port

    def _set_protocol_address(self, protocol, address):
        host, port = address
        self.set_config("dbms.connector.%s.listen_address" % protocol, "%s:%s" % (host, port))

    @property
    def http_address(self):
        """ The host and port on which this server expects HTTP communication.

        :returns: 2-tuple of (host, port)
        """
        return self._get_protocol_address("http", 7474)

    @http_address.setter
    def http_address(self, address):
        """ Set the host and port on which this server expects HTTP communication.
        """
        self._set_protocol_address("http", address)

    @property
    def https_address(self):
        """ The host and port on which this server expects HTTPS communication.

        :returns: 2-tuple of (host, port)
        """
        return self._get_protocol_address("https", 7473)

    @https_address.setter
    def https_address(self, address):
        """ Set the host and port on which this server expects HTTPS communication.
        """
        self._set_protocol_address("https", address)

    @property
    def bolt_address(self):
        """ The host and port on which this server expects Bolt communication.

        :returns: 2-tuple of (host, port)
        """
        return self._get_protocol_address("bolt", 7687)

    @bolt_address.setter
    def bolt_address(self, address):
        """ Set the host and port on which this server expects Bolt communication.
        """
        self._set_protocol_address("bolt", address)

    @property
    def http_uri(self):
        """ The full HTTP URI for this server.
        """
        host, port = self.http_address
        return "http://%s:%s" % (host, port)

    @property
    def https_uri(self):
        """ The full HTTPS URI for this server.
        """
        host, port = self.https_address
        return "https://%s:%s" % (host, port)

    @property
    def bolt_uri(self):
        """ The full Bolt URI for this server.
        """
        host, port = self.bolt_address
        return "bolt://%s:%s" % (host, port)

    @property
    def bolt_routing_uri(self):
        """ The full Bolt URI for this server.
        """
        host, port = self.bolt_address
        return "bolt+routing://%s:%s" % (host, port)

    def delete_store(self, force=False):
        """ Delete the store directory for this server.

        :param force:
        """
        if force or not self.server.running():
            try:
                rmtree(self.store_path, ignore_errors=force)
            except FileNotFoundError:
                pass
        else:
            raise RuntimeError("Refusing to drop database store while server is running")


class Server(object):
    """ Represents a Neo4j server process that can be started and stopped.
    """

    def __init__(self, installation):
        self.installation = installation

    @property
    def control_script(self):
        return path_join(self.installation.home, "bin", "neo4j")

    def start(self, wait=120.0, verbose=False):
        """ Start the server.
        """
        try:
            out = check_output("%s start" % self.control_script, shell=True)
        except CalledProcessError as error:
            if error.returncode == 2:
                raise OSError("Another process is listening on the server port")
            elif error.returncode == 512:
                raise OSError("Another server process is already running")
            else:
                raise OSError("An error occurred while trying to start "
                              "the server [%s]" % error.returncode)
        else:
            pid = None
            for line in out.decode("utf-8").splitlines(False):
                if verbose:
                    print(line)
                if line.startswith("process"):
                    number_in_brackets = re_compile("\[(\d+)\]")
                    numbers = number_in_brackets.search(line).groups()
                    if numbers:
                        pid = int(numbers[0])
                elif "(pid " in line:
                    pid = int(line.partition("(pid ")[-1].partition(")")[0])
            running = False
            address = self.installation.bolt_address or self.installation.http_address
            t0 = time()
            while not running and (time() - t0) < wait:
                try:
                    s = create_connection(address)
                except IOError:
                    sleep(0.5)
                else:
                    s.close()
                    running = True
            if not running:
                warn("Timed out waiting for server to start")
            return pid

    def stop(self):
        """ Stop the server.
        """
        pid = self.running()
        if not pid:
            return
        try:
            check_output(("%s stop" % self.control_script), shell=True)
        except CalledProcessError as error:
            raise OSError("An error occurred while trying to stop the server "
                          "[%s]" % error.returncode)
        while pid:
            try:
                kill(pid, 0)
            except OSError:
                pid = 0
            else:
                pass

    def restart(self):
        """ Restart the server.
        """
        self.stop()
        return self.start()

    def running(self):
        """ The PID of the current executing process for this server.
        """
        try:
            out = check_output(("%s status" % self.control_script), shell=True)
        except CalledProcessError as error:
            if error.returncode == 3:
                return None
            else:
                raise OSError("An error occurred while trying to query the "
                              "server status [%s]" % error.returncode)
        else:
            p = None
            for line in out.decode("utf-8").splitlines(False):
                if "running" in line:
                    p = int(line.rpartition(" ")[-1])
            return p


class AuthFile(object):
    """ A Neo4j auth file, generally located at data/dbms/auth.
    """

    def __init__(self, name):
        self.name = name
        if not isfile(self.name):
            d = dirname(self.name)
            try:
                makedirs(d)
            except OSError:
                pass
            with open(self.name, "wb"):
                pass

    def __iter__(self):
        with open(self.name, "rb") as f:
            for line in f:
                yield AuthUser.load(line)

    def remove(self, user_name):
        """ Remove a user.
        """
        user_name = bstr(user_name)
        with open(self.name, "rb") as f:
            lines = [line for line in f.readlines() if not AuthUser.match(line, user_name)]
        with open(self.name, "wb") as f:
            f.writelines(lines)

    def update(self, user_name, password):
        """ Add or update a user.
        """
        user_name = bstr(user_name)
        password = bstr(password)
        updated = False
        with open(self.name, "rb") as f:
            lines = []
            for line in f.readlines():
                if AuthUser.match(line, user_name):
                    lines.append(AuthUser.create(user_name, password).dump())
                    updated = True
                else:
                    lines.append(line)
        if not updated:
            lines.append(AuthUser.create(user_name, password).dump())
        with open(self.name, "wb") as f:
            f.writelines(lines)


class AuthUser(object):

    #: Name of user
    name = None

    #: The hash algorithm mused to encode the user data
    hash_algorithm = None

    #:
    digest = None

    #:
    salt = None

    @classmethod
    def create(cls, user_name, password):
        user_name = bstr(user_name)
        password = bstr(password)
        inst = cls(user_name, b"SHA-256", None, None)
        inst.set_password(password)
        return inst

    @classmethod
    def load(cls, s):
        s = bstr(s)
        fields = s.rstrip().split(b":")
        name = fields[0]
        hash_algorithm, digest, salt = fields[1].split(b",")
        return cls(name, hash_algorithm, unhex_bytes(digest), unhex_bytes(salt))

    @classmethod
    def match(cls, s, user_name):
        s = bstr(s)
        user_name = bstr(user_name)
        candidate_user_name, _, _ = s.partition(b":")
        return candidate_user_name == user_name

    def dump(self, eol=b"\r\n"):
        return self.name + b":" + self.hash_algorithm + b"," + hex_bytes(self.digest) + b"," + \
               hex_bytes(self.salt) + b":" + bstr(eol)

    def __init__(self, name, hash_algorithm, digest, salt):
        assert hash_algorithm == b"SHA-256"
        self.name = bstr(name)
        self.hash_algorithm = bstr(hash_algorithm)
        self.digest = digest
        self.salt = salt

    def __repr__(self):
        return "<AuthUser name=%r>" % self.name

    def set_password(self, password):
        assert self.hash_algorithm == b"SHA-256"
        salt = bytearray(randint(0x00, 0xFF) for _ in range(16))
        m = sha256()
        m.update(salt)
        m.update(bstr(password))
        self.digest = m.digest()
        self.salt = salt

    def check_password(self, password):
        assert self.hash_algorithm == b"SHA-256"
        m = sha256()
        m.update(self.salt)
        m.update(bstr(password))
        return m.digest() == self.digest



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from base64 import b64encode
from collections import OrderedDict
from json import dumps as json_dumps, loads as json_loads
from warnings import catch_warnings, simplefilter

from neo4j.addressing import SocketAddress
from neo4j.bolt import ServerInfo
from neo4j.exceptions import AuthError, Forbidden
from neo4j.v1 import Driver, Session, StatementResult, TransactionError, SessionError

from py2neo.data import Record
from py2neo.database import GraphError
from py2neo.internal.addressing import get_connection_data
from py2neo.internal.compat import urlsplit, ustr
from py2neo.internal.json import JSONDehydrator


# import logging
#
# # Enabling debugging at http.client level (requests->urllib3->http.client)
# # you will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
# # the only thing missing will be the response.body which is not logged.
# try: # for Python 3
#     from http.client import HTTPConnection
# except ImportError:
#     from httplib import HTTPConnection
# HTTPConnection.debuglevel = 1
#
# logging.basicConfig() # you need to initialize logging, otherwise you will not see anything from requests
# logging.getLogger().setLevel(logging.DEBUG)
# requests_log = logging.getLogger("urllib3")
# requests_log.setLevel(logging.DEBUG)
# requests_log.propagate = True


DEFAULT_PORT = 7474

OK = 200
CREATED = 201
NO_CONTENT = 204
UNAUTHORIZED = 401
FORBIDDEN = 403
NOT_FOUND = 404


def fix_parameters(parameters):
    if not parameters:
        return {}
    dehydrator = JSONDehydrator()
    try:
        dehydrated, = dehydrator.dehydrate([parameters])
    except TypeError as error:
        value = error.args[0]
        raise TypeError("Parameters of type {} are not supported".format(type(value).__name__))
    else:
        return dehydrated


class HTTP(object):
    """ Wrapper for HTTP method calls.
    """

    @staticmethod
    def authorization(user, password):
        return 'Basic ' + b64encode((user + ":" + password).encode("utf-8")).decode("ascii")

    def __init__(self, uri, headers, verified):
        self.uri = uri
        self.verified = verified
        parts = urlsplit(uri)
        scheme = parts.scheme
        host = parts.hostname
        port = parts.port
        if scheme == "http":
            from urllib3 import HTTPConnectionPool
            self._http = HTTPConnectionPool("%s:%d" % (host, port))
        elif scheme == "https":
            from urllib3 import HTTPSConnectionPool
            if verified:
                from certifi import where
                self._http = HTTPSConnectionPool("%s:%d" % (host, port), cert_reqs="CERT_REQUIRED", ca_certs=where())
            else:
                self._http = HTTPSConnectionPool("%s:%d" % (host, port))
        else:
            raise ValueError("Unsupported scheme %r" % scheme)
        self.path = parts.path
        if "auth" in headers:
            user, password = headers.pop("auth")
            headers["Authorization"] = 'Basic ' + b64encode(
                (ustr(user) + u":" + ustr(password)).encode("utf-8")).decode("ascii")
        self.headers = headers

    def __del__(self):
        self.close()

    def __eq__(self, other):
        try:
            return self.uri == other.uri
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def request(self, method, url, fields=None, headers=None, **urlopen_kw):
        from neo4j.v1 import ServiceUnavailable
        from urllib3.exceptions import MaxRetryError
        try:
            if self.verified:
                return self._http.request(method, url, fields, headers, **urlopen_kw)
            else:
                with catch_warnings():
                    simplefilter("ignore")
                    return self._http.request(method, url, fields, headers, **urlopen_kw)
        except MaxRetryError:
            raise ServiceUnavailable("Cannot send %s request to <%s>" % (method, url))

    def get_json(self, ref):
        """ Perform an HTTP GET to this resource and return JSON.
        """
        rs = self.request("GET", self.path + ref, headers=self.headers)
        try:
            if rs.status == 200:
                return json_loads(rs.data.decode('utf-8'))
            else:
                self.raise_error(rs.status, rs.data)
        finally:
            rs.close()

    def post(self, ref, json, expected):
        """ Perform an HTTP POST to this resource.
        """
        headers = dict(self.headers)
        if json is not None:
            headers["Content-Type"] = "application/json"
            json = json_dumps(json).encode('utf-8')
        rs = self.request("POST", self.path + ref, headers=self.headers, body=json)
        if rs.status not in expected:
            self.raise_error(rs.status, rs.data)
        return rs

    def delete(self, ref, expected):
        """ Perform an HTTP DELETE to this resource.
        """
        rs = self.request("DELETE", self.path + ref, headers=self.headers)
        if rs.status not in expected:
            self.raise_error(rs.status, rs.data)
        return rs

    def close(self):
        if self._http and self._http.pool:
            self._http.close()

    def raise_error(self, status_code, data):
        if status_code == UNAUTHORIZED:
            raise AuthError(self.uri)
        if status_code == FORBIDDEN:
            raise Forbidden(self.uri)
        if data:
            content = json_loads(data.decode('utf-8'))
        else:
            content = {}
        message = content.pop("message", "HTTP request to <%s> returned unexpected status code %s" % (self.uri, status_code))
        error = GraphError(message, **content)
        error.http_status_code = status_code
        raise error


class HTTPDriver(Driver):

    uri_scheme = "http"

    _connection_data = None

    _http = None

    _graph = None

    def __new__(cls, uri, **config):
        cls._check_uri(uri)
        instance = object.__new__(cls)
        instance._connection_data = connection_data = get_connection_data(uri, **config)
        instance._http = HTTP(connection_data["uri"] + "/db/data/", {
            "Authorization": HTTP.authorization(connection_data["user"], connection_data["password"]),
            "User-Agent": connection_data["user_agent"],
            "X-Stream": "true",
        }, connection_data["verified"])
        instance._graph = None
        return instance

    def session(self, access_mode=None, bookmark=None):
        if self._graph is None:
            from py2neo.database import Database
            self._graph = Database(self._connection_data["uri"], auth=self._connection_data["auth"]).default_graph
        return HTTPSession(self._graph, self._http)


class HTTPSDriver(Driver):

    uri_scheme = "https"

    _connection_data = None

    _http = None

    _graph = None

    def __new__(cls, uri, **config):
        cls._check_uri(uri)
        instance = object.__new__(cls)
        instance._connection_data = connection_data = get_connection_data(uri, **config)
        instance._http = HTTP(connection_data["uri"] + "/db/data/", {
            "Authorization": HTTP.authorization(connection_data["user"], connection_data["password"]),
            "User-Agent": connection_data["user_agent"],
            "X-Stream": "true",
        }, connection_data["verified"])
        instance._graph = None
        return instance

    def session(self, access_mode=None, bookmark=None):
        if self._graph is None:
            from py2neo.database import Database
            self._graph = Database(self._connection_data["uri"], auth=self._connection_data["auth"]).default_graph
        return HTTPSession(self._graph, self._http)


class HTTPResultLoader(object):

    def load(self, result):
        pass

    def fail(self):
        pass


class HTTPSession(Session):

    begin_ref = "transaction"

    autocommit_ref = "transaction/commit"

    transaction_ref = None      # e.g. "transaction/1"

    commit_ref = None           # e.g. "transaction/1/commit"

    def __init__(self, graph, http):
        self.graph = graph
        self.post = http.post
        self.delete = http.delete
        self.ref = self.autocommit_ref
        self._statements = []
        self._result_loaders = []
        self._bookmark = None

    def _connect(self, access_mode=None):
        pass

    def _disconnect(self, sync):
        if sync:
            self.sync()

    def run(self, statement, parameters=None, **kwparameters):
        if self.closed():
            raise SessionError("Session closed")
        if not statement:
            raise ValueError("Cannot run an empty statement")

        self._statements.append(OrderedDict([
            ("statement", ustr(statement)),
            ("parameters", fix_parameters(dict(parameters or {}, **kwparameters))),
            ("resultDataContents", ["REST"]),
            ("includeStats", True),
        ]))
        result_loader = HTTPResultLoader()
        self._result_loaders.append(result_loader)
        return HTTPStatementResult(self, result_loader)

    def send(self):
        self.sync()

    def fetch(self):
        return 0

    def sync(self):
        ref = self.ref
        # Some of the transactional URIs do not support empty statement
        # lists in versions earlier than 2.3. Which doesn't really matter
        # as it's a waste sending anything anyway.
        if ref in (self.autocommit_ref, self.begin_ref, self.transaction_ref) and not self._statements:
            return 0
        count = 0
        try:
            response = self.post(ref, {"statements": self._statements}, expected=(OK, CREATED))
            if response.status == 201:
                location_path = urlsplit(response.headers["Location"]).path
                self.transaction_ref = "".join(location_path.rpartition("transaction")[1:])
                self.commit_ref = "%s/commit" % self.transaction_ref
                self.ref = self.transaction_ref
            content = json_loads(response.data.decode("utf-8"))
            errors = content["errors"]
            if errors:
                from py2neo.database import GraphError
                raise GraphError.hydrate(errors[0])
            for i, result_loader in enumerate(self._result_loaders):
                try:
                    count += result_loader.load(content["results"][i])
                except IndexError:
                    result_loader.fail()
            return count
        finally:
            self._statements[:] = ()
            self._result_loaders[:] = ()

    def detach(self, result):
        return 0

    def last_bookmark(self):
        return None

    def commit_transaction(self):
        if not self.has_transaction():
            raise TransactionError("No transaction to commit")
        self._transaction = None
        self.ref = self.commit_ref or self.autocommit_ref
        try:
            self.sync()
        finally:
            self.commit_ref = self.transaction_ref = None
            self.ref = self.autocommit_ref
        self._bookmark = None
        return self._bookmark

    def rollback_transaction(self):
        if not self.has_transaction():
            raise TransactionError("No transaction to rollback")
        self._transaction = None
        self._bookmark = None
        try:
            if self.transaction_ref:
                self.ref = self.transaction_ref
                self.delete(self.ref, expected=(OK, NOT_FOUND))
        finally:
            self.commit_ref = self.transaction_ref = None
            self.ref = self.autocommit_ref

    def __begin__(self):
        self.ref = self.begin_ref


class HTTPStatementResult(StatementResult):

    zipper = Record

    def __init__(self, session, result_loader):
        from py2neo.internal.json import JSONHydrator

        super(HTTPStatementResult, self).__init__(session, JSONHydrator(session.graph, ()))

        def load(result):
            from neo4j.v1 import BoltStatementResultSummary

            keys = self._keys = self._hydrant.keys = tuple(result["columns"])
            hydrate = self._hydrant.hydrate
            for record in result["data"]:
                self._records.append(Record(zip(keys, hydrate(record["rest"]))))

            stats = result["stats"]
            # fix broken key
            if "relationship_deleted" in stats:
                stats["relationships_deleted"] = stats["relationship_deleted"]
                del stats["relationship_deleted"]
            if "contains_updates" in stats:
                del stats["contains_updates"]

            metadata = {"stats": stats}
            if "plan" in result:
                metadata["http_plan"] = result["plan"]

            # TODO: fill these in
            metadata["statement"] = None
            metadata["parameters"] = None

            metadata["server"] = ServerInfo(SocketAddress.from_uri(session.graph.database.uri))

            self._summary = BoltStatementResultSummary(**metadata)
            self._session = None
            return len(self._records)

        def fail():
            self._session = None

        result_loader.load = load
        result_loader.fail = fail

    def records(self):
        """ Generator for records obtained from this result.

        :yields: iterable of :class:`.Record` objects
        """
        if self.attached():
            self._session.send()
        records = self._records
        next_record = records.popleft
        while records:
            yield next_record()
